import numpy as np
from numpy.linalg import solve, norm

import matplotlib.pyplot as plt

def savefig(basename):
    plt.savefig("img/" + basename + ".pdf")
    plt.savefig("img/" + basename + ".png")

###################
# Unteraufgabe a) #
###################

def integrate(f, Jf, y0, dt, n_steps, row_step):
    n_vars = y0.shape[0]
    t = np.arange(n_steps+1)*dt
    y = np.empty((n_steps+1, n_vars))

    y[0,:] = y0
    for i in range(n_steps):
        y[i+1,:] = row_step(f, Jf, y[i,:], dt)

    return t, y

def row_2_step(f, Jf, y0, dt):
    """Rosenbrock-Wanner Methode der Ordnung 2

    Input:
         f : Die rechte Seite der ODE f(x).

        Jf : Jacobi Matrix J(x) der Funktion, `shape == (n, n)`.

        y0 : ndarray.
             Aktueller Wert der approximativen Lösung der ODE.

        dt : Schrittweite

    Output:
        y1 : Zeitpropagierter Wert y(t+h).
    """
    # TODO implement ROW-2
    return np.zero_like(y0)

def row_2(f, Jf, y0, dt, n_steps):
    return integrate(f, Jf, y0, dt, n_steps, row_2_step)

def row_3_step(f, Jf, y0, dt):
    """Rosenbrock-Wanner Methode der Ordnung 3

    Input:
         f : Die rechte Seite der ODE f(x).

        Jf : Jacobi Matrix J(x) der Funktion, `shape == (n, n)`.

        y0 : ndarray.
             Aktueller Wert der approximativen Lösung der ODE.

        dt : Schrittweite

    Output:
        y1 : Zeitpropagierter Wert y(t+h).
    """
    # TODO implement ROW-3
    return np.zero_like(y0)

def row_3(f, Jf, y0, dt, n_steps):
    return integrate(f, Jf, y0, dt, n_steps, row_3_step)


###################
# Unteraufgabe b) #
###################

def aufgabe_b():
    print(" Aufgabe b)")
    # TODO implement b)

    t2, s2 = row_2(f, Jf, y0, h, N)
    t3, s3 = row_3(f, Jf, y0, h, N)

    t = np.linspace(0.0, T, N+1)
    y = np.ones_like(s2)  # TODO set this to the exact solution

    plt.figure()
    plt.plot(t, y, "-d", label="Exact $y(t)$, $\lambda = %3.1f$" % l)
    plt.plot(t2, s2, "-s", label="ROW-2 $y(t)$, $\lambda = %3.1f$" % l)
    plt.plot(t3, s3, "-o", label="ROW-3 $y(t)$, $\lambda = %3.1f$" % l)
    plt.grid(True)
    plt.xlim(t.min(), t.max())
    plt.ylim(0, 1.05)
    plt.legend(loc="lower right")
    plt.xlabel(r"Time $t$")
    plt.ylabel(r"Solution $y(t)$")
    savefig("solution_logistic_row")

    plt.figure()
    plt.plot(t, s2 - y, "-bo", label="ROW-2")
    plt.plot(t, s3 - y, "-gs", label="ROW-3")
    plt.grid(True)
    plt.xlim(t.min(), t.max())
    plt.legend(loc="upper right")
    plt.xlabel(r"Time $t$")
    plt.ylabel(r"Absolute Error")
    savefig("error_logistic_row")


###################
# Unteraufgabe c) #
###################

def aufgabe_c():
    print(" Aufgabe c)")

    # TODO set up the problem

    all_n_steps = 2**np.arange(4,13)
    error_row_2 = np.empty(all_n_steps.size)
    error_row_3 = np.empty_like(error_row_2)

    # TODO implement error loop.

    plt.figure()
    plt.loglog(all_n_steps, error_row_2, "b-o", label="ROW-2")
    plt.loglog(all_n_steps, error_row_3, "g-o", label="ROW-3")
    plt.loglog(all_n_steps, 1e-5*(T/all_n_steps)**2, "-k", label="$O(N^-2)$")
    plt.loglog(all_n_steps, 1e-5*(T/all_n_steps)**3, "--k", label="$O(N^-3)$")
    plt.grid(True)
    plt.xlim(all_n_steps.min(), all_n_steps.max())
    plt.legend(loc="lower left")
    plt.xlabel(r"Number of steps $N$")
    plt.ylabel(r"Absolute Error at $T = %.1f$" % T)
    savefig("convergence_row")


###################
# Unteraufgabe d) #
###################

def odeintadapt(Psilow, Psihigh, t_end, y0, fy0=None, dt0=None, dtmin=None, reltol=1e-2, abstol=1e-4):
    """Adaptive Integrator for stiff ODEs and Systems
    based on two suitable methods of different order.

    Input:
    Psilow:   Integrator of low order
    Psihigh:  Integrator of high order
    t_end:    Endtime
    y0:       Initial value, ndarray with shape == (n,).
    fy0:      The value f(y0): R^n used to estimate initial timestep size
    dt0:       Initial timestep (optional)
    dtmin:     Minimal timestep (optional)
    reltol:   Relative tolerance (optional)
    abstol:   Absolute Tolerance (optional)

    Output:
           t : Accepted time-steps
           y : Solution at the time-steps
    rejected : Array of rejected time-steps.
    error_estimates : Estimated error for each accepted time-step.
    """
    # Heuristic choice of initial timestep size
    if dt0 is None:
        dt0 = t_end / (100.0*(norm(fy0) + 0.1))
    if dtmin is None:
        dtmin = dt0 / 10000.0

    # Initial values
    yi = np.copy(y0)
    ti = 0.0
    dt = dt0

    # Use dynamically growing memory, since the algorithm
    # is adaptive.
    t = [ti]
    y = [yi]
    rejected = []
    error_estimates = [0.0]

    # TODO implement the adaptive time integration loop.

    n = yi.shape[0]
    t = np.array(t).reshape(-1)
    y = np.array(y).reshape(-1, n)
    rejected = np.array(rejected)
    error_estimates = np.array(error_estimates)
    return t, y, rejected, error_estimates


###################
# Unteraufgabe e) #
###################

def aufgabe_e():
    print(" Aufgabe e)")

    # TODO set up e)

    # Discrete Evolution Operators
    Psilow = lambda dt, y: row_2_step(f, Jf, y, dt)
    Psihigh = lambda dt, y: row_3_step(f, Jf, y, dt)

    # Adaptive timestepping
    ta, ya, _, _ = odeintadapt(Psilow, Psihigh, T, y0, f(y0))
    nsteps = ta.size

    print("Es werden %d Zeitschritte benoetigt" % nsteps)

    n_steps = 100
    t2, s2 = row_2(f, Jf, y0, T/float(n_steps), n_steps)
    t3, s3 = row_3(f, Jf, y0, T/float(n_steps), n_steps)

    plt.figure()
    plt.plot(t2, s2, "-bd", label=r"ROW-2 $y(t)$, $\lambda = %3.1f$" % l)
    plt.plot(t3, s3, "-gs", label=r"ROW-3 $y(t)$, $\lambda = %3.1f$" % l)
    plt.plot(ta, ya, "-ro", label=r"Adaptive $y(t)$, $\lambda = %3.1f$" % l)
    plt.grid(True)
    plt.xlim(0, T)
    plt.ylim(0, 1.05)
    plt.legend(loc="best")
    plt.xlabel(r"Time $t$")
    plt.ylabel(r"Solution $y(t)$")
    savefig("solution_logistic_adaptive")

    plt.figure()
    plt.plot(t2, s2 - ode.exact(y0, t2), "-bd", label=r"ROW-2 $y(t)$")
    plt.plot(t3, s3 - ode.exact(y0, t3), "-gs", label=r"ROW-3 $y(t)$")
    plt.plot(ta, ya - ode.exact(y0, ta), "-ro", label=r"Adaptive $y(t)$")
    plt.grid(True)
    plt.xlim(0, T)
    plt.ylim(-0.0025, 0.0025)
    plt.legend(loc="best")
    plt.xlabel(r"Time $t$")
    plt.ylabel(r"Absolute Error")
    savefig("error_logistic_adaptive")


###################
# Unteraufgabe f) #
###################

def aufgabe_f():
    print(" Aufgabe f)")

    # TODO set up problem

    # Discrete Evolution Operators
    Psilow =  lambda h,y: row_2_step(f, Jf, y, h)
    Psihigh = lambda h,y: row_3_step(f, Jf, y, h)

    # Adaptive timestepping
    ta, ya, _, _ = odeintadapt(Psilow, Psihigh, T, y0, dt0=0.1)

    plt.figure()
    plt.plot(ta, ya[:,0], "-+", label=r"$y_0(t)$")
    plt.plot(ta, ya[:,1], "-+", label=r"$y_1(t)$")
    plt.grid(True)
    plt.xlim(0, T)
    plt.legend(loc="upper right")
    plt.xlabel(r"Time $t$")
    plt.ylabel(r"Solution $y(t)$")
    savefig("solution_system")


###################
# Unteraufgabe g) #
###################

def aufgabe_g():
    print(" Aufgabe g)")

    # TODO set up g)

    # Discrete Evolution Operators
    Psilow =  lambda h,y: row_2_step(f, Jf, y, h)
    Psihigh = lambda h,y: row_3_step(f, Jf, y, h)

    # Adaptive timestepping
    ta, ya, _, _ = odeintadapt(Psilow, Psihigh, T, y0, f(y0))

    hmin = np.min(np.diff(ta))
    n_steps = ta.size

    plt.figure()
    plt.plot(ta, ya, "+-")
    plt.grid(True)
    plt.xlim(0, T)
    plt.ylim(0, 1.1)
    plt.xlabel(r"Time $t$")
    plt.ylabel(r"Solution $y(t)$")
    savefig("solution_superstiff")

    plt.figure()
    plt.plot(ta[:-1], np.diff(ta), "-o")
    plt.grid(True)
    plt.xlim(0, T)
    plt.xlabel(r"Time $t$")
    plt.xlabel(r"Timestep size $h_j$")
    savefig("timestepsize_superstiff")

    print("Minimal steps size: %f" % hmin)
    print("Number of adaptive steps: %i" % n_steps)
    print("Number of non-adaptive steps: %.2f" % (T/hmin))




if __name__ == "__main__":
    # aufgabe_b()
    # aufgabe_c()
    # aufgabe_e()
    # aufgabe_f()
    # aufgabe_g()
